require 'openssl'
require 'base64'

module UU
  module OS
    module Security

      # Provides data encryption and decryption.
      #
      # @example Basic usage - +include+ allows to use +Encryption+ methods in instance methods of a class.
      #   require 'uu/os/security/encryption'
      #
      #   class Foo
      #     include UU::OS::Security::Encryption
      #
      #     def bar(data)
      #       encrypted_data = encrypt(data)
      #     end
      #
      #     def baz(data)
      #       decrypted_data = decrypt(data)
      #     end
      #   end
      #
      # @example Usage with class methods (singleton methods) - +extend+ allows to use +Encryption+ methods in class methods (but not in instance methods).
      #   require 'uu/os/security/encryption'
      #
      #   class Foo
      #     extend UU::OS::Security::Encryption
      #
      #     def self.bar(data)
      #       encrypted_data = encrypt(data)
      #     end
      #
      #     def self.baz(data)
      #       decrypted_data = decrypt(data)
      #     end
      #   end
      #
      # @example Usage in a module - +extend+ allows to use +Encryption+ methods in module methods.
      #   require 'uu/os/security/encryption'
      #
      #   module Foo
      #     extend UU::OS::Security::Encryption
      #
      #     def bar(data)
      #       encrypted_data = encrypt(data)
      #     end
      #
      #     def baz(data)
      #       decrypted_data = decrypt(data)
      #     end
      #   end
      #
      # @example Usage with both class methods (singleton methods) and instance methods.
      #   require 'uu/os/security/encryption'
      #
      #   class Foo
      #     include UU::OS::Security::Encryption
      #     extend UU::OS::Security::Encryption
      #
      #     def self.bar(data)
      #       encrypted_data = encrypt(data)
      #     end
      #
      #     def baz(data)
      #       decrypted_data = decrypt(data)
      #     end
      #   end
      module Encryption

        # Encrypts the given data.
        #
        # @param data [String] The data to encrypt.
        # @param password [String] Password to be used for encryption. In case no
        #    password is provided, default one is used meaning that decryption can
        #    be done by anybody using {#decrypt decrypt} method.
        # @return [String] The encrypted data.
        def encrypt(data, password = nil)
          salt = ''
          8.times {salt << rand(255).chr}
          cipher = OpenSSL::Cipher.new(ALGORITHM)
          cipher.encrypt
          key = OpenSSL::PKCS5.pbkdf2_hmac_sha1(password.nil? ? SECRET : password, salt, 1000, 128)          
          cipher.key = key
          cipher.iv = key
          tmp = cipher.update(data) + cipher.final
          Base64.strict_encode64("Salted__#{salt}#{tmp}")
        end

        # Decrypts the given data.
        #
        # @param data [String] The data to decrypt.
        # @param password [String] Password to be used for decryption (must be same
        #    as password for encryption, else decryption fails). In case no password
        #    is provided, default one is used.
        # @return [String] The decrypted data.
        def decrypt(data, password = nil)
          data = Base64.decode64(data)
          salt = data[8..15]
          data = data[16..-1]
          cipher = OpenSSL::Cipher.new(ALGORITHM)
          cipher.decrypt
          begin
            key = OpenSSL::PKCS5.pbkdf2_hmac_sha1(password.nil? ? SECRET : password, salt, 1000, 128)
            cipher.key = key
            cipher.iv = key
            cipher.update(data) + cipher.final
          rescue
            cipher.reset
            cipher.pkcs5_keyivgen(SECRET, salt , 1)
            cipher.update(data) + cipher.final
          end
        end

        private

        # Default cipher algorithm.
        # @private
        # Note: 128-bit key was chosen because it does not require to install
        # Java Cryptography Extenstion Unlimited Jurisdiction Policy Files
        # when using JRuby. We don't want to force users to install it since
        # this requirement would make the uuADK client installation more
        # complicated. However, this may change in the future (e.g. when
        # a client installer is provided for the uuADK).
        ALGORITHM = 'AES-128-CBC'

        # The secret phrase used as password for the cipher.
        # @private
        SECRET = 'U2FsdGVkX19jrg+FR+SgV5nG6ulKFh+SvsegRXSV5kc=\n'.freeze

      end

    end
  end
end
